@@ -72,6 +72,24 @@
  *
  * A same #GtkSourceCompletionProvider object can be used for several
  * #GtkSourceCompletion.
+ *
+ * # GtkSourceCompletion as GtkBuildable
+ *
+ * The GtkSourceCompletion implementation of the #GtkBuildable interface exposes
+ * the info window object (see gtk_source_completion_get_info_window()) with the
+ * internal-child "info_window".
+ *
+ * An example of a UI definition fragment with GtkSourceCompletion:
+ * |[
+ * <object class="GtkSourceCompletion">
+ *   <property name="select_on_show">False</property>
+ *   <child internal-child="info_window">
+ *     <object class="GtkSourceCompletionInfo">
+ *       <property name="border_width">6</property>
+ *     </object>
+ *   </child>
+ * </object>
+ * ]|
  */
 
 /* Idea to improve the code: use a composite widget template. This class is not
@@ -89,7 +107,6 @@
 #include "gtksourcecompletioncontainer.h"
 #include "gtksourcebuffer.h"
 #include "gtksourceview.h"
-#include "gtksourceview-marshal.h"
 #include "gtksourceview-i18n.h"
 
 /* Signals */
@@ -121,13 +138,6 @@ enum
 	PROP_PROPOSAL_PAGE_SIZE
 };
 
-enum
-{
-	TEXT_BUFFER_DELETE_RANGE,
-	TEXT_BUFFER_INSERT_TEXT,
-	LAST_EXTERNAL_SIGNAL
-};
-
 struct _GtkSourceCompletionPrivate
 {
 	GtkSourceCompletionInfo *main_window;
@@ -165,13 +175,19 @@ struct _GtkSourceCompletionPrivate
 
 	guint show_timed_out_id;
 
-	gulong signals_ids[LAST_EXTERNAL_SIGNAL];
+	GtkTextBuffer *buffer;
 
 	GList *auto_completion_selection;
 	GtkSourceCompletionContext *auto_completion_context;
 
+	/* Number of times the interactive completion is blocked */
+	guint block_interactive_num;
+
 	/* Properties */
 
+	/* Weak reference to the view. You must check if view != NULL before
+	 * using it.
+	 */
 	GtkSourceView *view;
 	guint num_accelerators;
 	guint auto_complete_delay;
@@ -186,7 +202,12 @@ struct _GtkSourceCompletionPrivate
 
 static guint signals[LAST_SIGNAL] = { 0 };
 
-G_DEFINE_TYPE_WITH_PRIVATE (GtkSourceCompletion, gtk_source_completion, G_TYPE_OBJECT)
+static void	gtk_source_completion_buildable_interface_init (GtkBuildableIface *iface);
+
+G_DEFINE_TYPE_WITH_CODE (GtkSourceCompletion, gtk_source_completion, G_TYPE_OBJECT,
+			 G_ADD_PRIVATE (GtkSourceCompletion)
+			 G_IMPLEMENT_INTERFACE (GTK_TYPE_BUILDABLE,
+						gtk_source_completion_buildable_interface_init))
 
 static void
 scroll_to_iter (GtkSourceCompletion *completion,
@@ -291,12 +312,9 @@ static void
 get_iter_at_insert (GtkSourceCompletion *completion,
                     GtkTextIter         *iter)
 {
-	GtkTextBuffer *buffer;
-
-	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (completion->priv->view));
-	gtk_text_buffer_get_iter_at_mark (buffer,
+	gtk_text_buffer_get_iter_at_mark (completion->priv->buffer,
 	                                  iter,
-	                                  gtk_text_buffer_get_insert (buffer));
+	                                  gtk_text_buffer_get_insert (completion->priv->buffer));
 }
 
 static GList *
@@ -436,6 +454,11 @@ update_window_position (GtkSourceCompletion *completion)
 	GtkTextIter iter;
 	gboolean iter_set = FALSE;
 
+	if (completion->priv->view == NULL)
+	{
+		return;
+	}
+
 	/* The model can be modified while there is no completion context, for
 	 * example when the headers are shown or hidden. This triggers a signal
 	 * to update the window position, but if there is no completion context,
@@ -465,9 +488,7 @@ update_window_position (GtkSourceCompletion *completion)
 	if (!iter_set)
 	{
 		GtkTextIter end_word;
-		GtkTextBuffer *buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (completion->priv->view));
-
-		get_word_iter (buffer, &iter, &end_word);
+		get_word_iter (completion->priv->buffer, &iter, &end_word);
 	}
 
 	gtk_source_completion_info_move_to_iter (completion->priv->main_window,
@@ -564,8 +585,18 @@ update_proposal_info (GtkSourceCompletion *completion)
 static void
 gtk_source_completion_show_default (GtkSourceCompletion *completion)
 {
+	if (completion->priv->view == NULL)
+	{
+		return;
+	}
+
 	gtk_widget_show (GTK_WIDGET (completion->priv->main_window));
 
+	/* Do the autosize when the widget is visible. It doesn't work if it is
+	 * done before.
+	 */
+	gtk_tree_view_columns_autosize (completion->priv->tree_view_proposals);
+
 	if (!completion->priv->remember_info_visibility)
 	{
 		gtk_toggle_button_set_active (completion->priv->info_button, FALSE);
@@ -637,6 +668,11 @@ gtk_source_completion_activate_proposal (GtkSourceCompletion *completion)
 	GtkTextIter insert_iter;
 	gboolean activated;
 
+	if (completion->priv->view == NULL)
+	{
+		return;
+	}
+
 	if (!get_selected_proposal (completion, &provider, &proposal))
 	{
 		return;
@@ -651,7 +687,6 @@ gtk_source_completion_activate_proposal (GtkSourceCompletion *completion)
 	if (!activated)
 	{
 		GtkTextIter start_iter;
-		GtkTextBuffer *buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (completion->priv->view));
 		gchar *text = gtk_source_completion_proposal_get_text (proposal);
 
 		gboolean has_start = gtk_source_completion_provider_get_start_iter (provider,
@@ -661,14 +696,14 @@ gtk_source_completion_activate_proposal (GtkSourceCompletion *completion)
 
 		if (has_start)
 		{
-			gtk_text_buffer_begin_user_action (buffer);
-			gtk_text_buffer_delete (buffer, &start_iter, &insert_iter);
-			gtk_text_buffer_insert (buffer, &start_iter, text, -1);
-			gtk_text_buffer_end_user_action (buffer);
+			gtk_text_buffer_begin_user_action (completion->priv->buffer);
+			gtk_text_buffer_delete (completion->priv->buffer, &start_iter, &insert_iter);
+			gtk_text_buffer_insert (completion->priv->buffer, &start_iter, text, -1);
+			gtk_text_buffer_end_user_action (completion->priv->buffer);
 		}
 		else
 		{
-			replace_current_word (buffer, text);
+			replace_current_word (completion->priv->buffer, text);
 		}
 
 		g_free (text);
@@ -1560,7 +1595,15 @@ gtk_source_completion_dispose (GObject *object)
 
 	reset_completion (completion);
 
-	g_clear_object (&completion->priv->view);
+	if (completion->priv->view != NULL)
+	{
+		g_object_remove_weak_pointer (G_OBJECT (completion->priv->view),
+					      (gpointer *)&completion->priv->view);
+
+		completion->priv->view = NULL;
+	}
+
+	g_clear_object (&completion->priv->buffer);
 	g_clear_object (&completion->priv->default_info);
 	g_clear_object (&completion->priv->model_proposals);
 
@@ -1582,11 +1625,132 @@ gtk_source_completion_dispose (GObject *object)
 	G_OBJECT_CLASS (gtk_source_completion_parent_class)->dispose (object);
 }
 
+/* Unconditionnally block interactive completion, without taking into account
+ * priv->block_interactive_num.
+ * g_signal_handlers_block_by_func() has a counter too, so you may think that
+ * block_interactive_num is useless. But it is useful when the buffer changes,
+ * to keep the signal handler blocked on the new buffer.
+ */
 static void
-connect_view (GtkSourceCompletion *completion)
+block_interactive (GtkSourceCompletion *completion)
 {
-	GtkTextBuffer *buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (completion->priv->view));
-	GtkStyleContext *style_context = gtk_widget_get_style_context (GTK_WIDGET (completion->priv->view));
+	g_signal_handlers_block_by_func (completion->priv->buffer,
+					 buffer_insert_text_cb,
+					 completion);
+
+	g_signal_handlers_block_by_func (completion->priv->buffer,
+					 buffer_delete_range_cb,
+					 completion);
+}
+
+static void
+connect_buffer (GtkSourceCompletion *completion)
+{
+	GtkTextBuffer *new_buffer = NULL;
+
+	if (completion->priv->view != NULL)
+	{
+		new_buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (completion->priv->view));
+	}
+
+	if (completion->priv->buffer == new_buffer)
+	{
+		return;
+	}
+
+	if (completion->priv->buffer != NULL)
+	{
+		g_signal_handlers_disconnect_by_func (completion->priv->buffer,
+						      buffer_mark_set_cb,
+						      completion);
+
+		g_signal_handlers_disconnect_by_func (completion->priv->buffer,
+						      gtk_source_completion_block_interactive,
+						      completion);
+
+		g_signal_handlers_disconnect_by_func (completion->priv->buffer,
+						      gtk_source_completion_unblock_interactive,
+						      completion);
+
+		g_signal_handlers_disconnect_by_func (completion->priv->buffer,
+						      buffer_delete_range_cb,
+						      completion);
+
+		g_signal_handlers_disconnect_by_func (completion->priv->buffer,
+						      buffer_insert_text_cb,
+						      completion);
+
+		reset_completion (completion);
+
+		g_object_unref (completion->priv->buffer);
+	}
+
+	completion->priv->buffer = new_buffer;
+
+	if (new_buffer == NULL)
+	{
+		return;
+	}
+
+	g_object_ref (completion->priv->buffer);
+
+	g_signal_connect_object (new_buffer,
+				 "mark-set",
+				 G_CALLBACK (buffer_mark_set_cb),
+				 completion,
+				 G_CONNECT_AFTER);
+
+	g_signal_connect_object (new_buffer,
+		                 "undo",
+		                 G_CALLBACK (gtk_source_completion_block_interactive),
+		                 completion,
+		                 G_CONNECT_SWAPPED);
+
+	g_signal_connect_object (new_buffer,
+		                 "undo",
+		                 G_CALLBACK (gtk_source_completion_unblock_interactive),
+		                 completion,
+		                 G_CONNECT_SWAPPED | G_CONNECT_AFTER);
+
+	g_signal_connect_object (new_buffer,
+		                 "redo",
+		                 G_CALLBACK (gtk_source_completion_block_interactive),
+		                 completion,
+		                 G_CONNECT_SWAPPED);
+
+	g_signal_connect_object (new_buffer,
+		                 "redo",
+		                 G_CALLBACK (gtk_source_completion_unblock_interactive),
+		                 completion,
+		                 G_CONNECT_SWAPPED | G_CONNECT_AFTER);
+
+	g_signal_connect_object (new_buffer,
+				 "delete-range",
+				 G_CALLBACK (buffer_delete_range_cb),
+				 completion,
+				 G_CONNECT_AFTER);
+
+	g_signal_connect_object (new_buffer,
+				 "insert-text",
+				 G_CALLBACK (buffer_insert_text_cb),
+				 completion,
+				 G_CONNECT_AFTER);
+
+	if (completion->priv->block_interactive_num > 0)
+	{
+		block_interactive (completion);
+	}
+}
+
+static void
+connect_view (GtkSourceCompletion *completion,
+	      GtkSourceView       *view)
+{
+	g_assert (completion->priv->view == NULL);
+	completion->priv->view = view;
+
+	g_object_add_weak_pointer (G_OBJECT (view),
+				   (gpointer *)&completion->priv->view);
 
 	g_signal_connect_object (completion->priv->view,
 				 "focus-out-event",
@@ -1618,57 +1782,13 @@ connect_view (GtkSourceCompletion *completion)
 				 completion,
 				 G_CONNECT_SWAPPED | G_CONNECT_AFTER);
 
-	g_signal_connect_object (buffer,
-				 "mark-set",
-				 G_CALLBACK (buffer_mark_set_cb),
+	connect_buffer (completion);
+
+	g_signal_connect_object (completion->priv->view,
+				 "notify::buffer",
+				 G_CALLBACK (connect_buffer),
 				 completion,
-				 G_CONNECT_AFTER);
-
-	g_signal_connect_object (buffer,
-		                 "undo",
-		                 G_CALLBACK (gtk_source_completion_block_interactive),
-		                 completion,
-		                 G_CONNECT_SWAPPED);
-
-	g_signal_connect_object (buffer,
-		                 "undo",
-		                 G_CALLBACK (gtk_source_completion_unblock_interactive),
-		                 completion,
-		                 G_CONNECT_SWAPPED | G_CONNECT_AFTER);
-
-	g_signal_connect_object (buffer,
-		                 "redo",
-		                 G_CALLBACK (gtk_source_completion_block_interactive),
-		                 completion,
-		                 G_CONNECT_SWAPPED);
-
-	g_signal_connect_object (buffer,
-		                 "redo",
-		                 G_CALLBACK (gtk_source_completion_unblock_interactive),
-		                 completion,
-		                 G_CONNECT_SWAPPED | G_CONNECT_AFTER);
-
-	completion->priv->signals_ids[TEXT_BUFFER_DELETE_RANGE] =
-		g_signal_connect_object (buffer,
-					 "delete-range",
-					 G_CALLBACK (buffer_delete_range_cb),
-					 completion,
-					 G_CONNECT_AFTER);
-
-	completion->priv->signals_ids[TEXT_BUFFER_INSERT_TEXT] =
-		g_signal_connect_object (buffer,
-					 "insert-text",
-					 G_CALLBACK (buffer_insert_text_cb),
-					 completion,
-					 G_CONNECT_AFTER);
-
-	g_signal_connect_object (style_context,
-				 "changed",
-				 G_CALLBACK (style_context_changed),
-				 completion,
-				 G_CONNECT_AFTER);
-
-	style_context_changed (style_context, completion);
+				 G_CONNECT_SWAPPED);
 }
 
 static void
@@ -1733,9 +1853,7 @@ gtk_source_completion_set_property (GObject      *object,
 	switch (prop_id)
 	{
 		case PROP_VIEW:
-			/* On construction only */
-			completion->priv->view = g_value_dup_object (value);
-			connect_view (completion);
+			connect_view (completion, g_value_get_object (value));
 			break;
 		case PROP_REMEMBER_INFO_VISIBILITY:
 			completion->priv->remember_info_visibility = g_value_get_boolean (value);
@@ -1773,6 +1891,273 @@ gtk_source_completion_set_property (GObject      *object,
 	}
 }
 
+static gboolean
+selection_func (GtkTreeSelection    *selection,
+                GtkTreeModel        *model,
+                GtkTreePath         *path,
+                gboolean             path_currently_selected,
+                GtkSourceCompletion *completion)
+{
+	GtkTreeIter iter;
+
+	gtk_tree_model_get_iter (model, &iter, path);
+
+	if (gtk_source_completion_model_iter_is_header (completion->priv->model_proposals,
+	                                                &iter))
+	{
+		/* A header must never be selected */
+		g_return_val_if_fail (!path_currently_selected, TRUE);
+		return FALSE;
+	}
+	else
+	{
+		return TRUE;
+	}
+}
+
+static void
+accelerators_notify_cb (GtkSourceCompletion *completion,
+			GParamSpec          *pspec,
+			GtkTreeViewColumn   *column)
+{
+	gtk_tree_view_column_set_visible (column, completion->priv->num_accelerators > 0);
+}
+
+static void
+init_tree_view (GtkSourceCompletion *completion,
+		GtkBuilder          *builder)
+{
+	GtkTreeSelection *selection;
+	GtkTreeViewColumn *column;
+	GtkCellRenderer *cell_renderer;
+	GtkStyleContext *style_context;
+	GdkRGBA background_color;
+	GdkRGBA foreground_color;
+
+	completion->priv->tree_view_proposals = GTK_TREE_VIEW (gtk_builder_get_object (builder, "tree_view_proposals"));
+
+	g_signal_connect_swapped (completion->priv->tree_view_proposals,
+				  "row-activated",
+				  G_CALLBACK (gtk_source_completion_activate_proposal),
+				  completion);
+
+	g_signal_connect_swapped (completion->priv->tree_view_proposals,
+				  "size-allocate",
+				  G_CALLBACK (gtk_source_completion_proposals_size_allocate),
+				  completion);
+
+	/* Selection */
+
+	selection = gtk_tree_view_get_selection (completion->priv->tree_view_proposals);
+
+	gtk_tree_selection_set_select_function (selection,
+	                                        (GtkTreeSelectionFunc)selection_func,
+	                                        completion,
+	                                        NULL);
+
+	g_signal_connect (selection,
+			  "changed",
+			  G_CALLBACK (selection_changed_cb),
+			  completion);
+
+	/* Icon cell renderer */
+
+	cell_renderer = GTK_CELL_RENDERER (gtk_builder_get_object (builder, "cell_renderer_icon"));
+
+	column = GTK_TREE_VIEW_COLUMN (gtk_builder_get_object (builder, "tree_view_column_icon"));
+
+	gtk_tree_view_column_set_attributes (column, cell_renderer,
+					     "pixbuf", GTK_SOURCE_COMPLETION_MODEL_COLUMN_ICON,
+					     "cell-background-set", GTK_SOURCE_COMPLETION_MODEL_COLUMN_IS_HEADER,
+					     NULL);
+
+	style_context = gtk_widget_get_style_context (GTK_WIDGET (completion->priv->tree_view_proposals));
+	gtk_style_context_get_background_color (style_context,
+	                                        GTK_STATE_FLAG_INSENSITIVE,
+	                                        &background_color);
+
+	g_object_set (cell_renderer,
+	              "cell-background-rgba", &background_color,
+	              NULL);
+
+	g_object_bind_property (completion, "show-icons",
+				cell_renderer, "visible",
+				G_BINDING_SYNC_CREATE);
+
+	/* Proposal text cell renderer */
+
+	cell_renderer = GTK_CELL_RENDERER (gtk_builder_get_object (builder, "cell_renderer_proposal"));
+	completion->priv->cell_renderer_proposal = cell_renderer;
+
+	column = GTK_TREE_VIEW_COLUMN (gtk_builder_get_object (builder, "tree_view_column_proposal"));
+
+	gtk_tree_view_column_set_attributes (column, cell_renderer,
+					     "markup", GTK_SOURCE_COMPLETION_MODEL_COLUMN_MARKUP,
+					     "cell-background-set", GTK_SOURCE_COMPLETION_MODEL_COLUMN_IS_HEADER,
+					     "foreground-set", GTK_SOURCE_COMPLETION_MODEL_COLUMN_IS_HEADER,
+					     NULL);
+
+	gtk_style_context_get_color (style_context,
+	                             GTK_STATE_FLAG_INSENSITIVE,
+	                             &foreground_color);
+
+	g_object_set (cell_renderer,
+	              "foreground-rgba", &foreground_color,
+	              "cell-background-rgba", &background_color,
+	              NULL);
+
+	/* Accelerators cell renderer */
+
+	column = GTK_TREE_VIEW_COLUMN (gtk_builder_get_object (builder, "tree_view_column_accelerator"));
+
+	cell_renderer = GTK_CELL_RENDERER (gtk_builder_get_object (builder, "cell_renderer_accelerator"));
+
+	gtk_tree_view_column_set_attributes (column,
+					     cell_renderer,
+					     "cell-background-set", GTK_SOURCE_COMPLETION_MODEL_COLUMN_IS_HEADER,
+					     NULL);
+
+	g_object_set (cell_renderer,
+	              "foreground-rgba", &foreground_color,
+	              "cell-background-rgba", &background_color,
+		      NULL);
+
+	gtk_tree_view_column_set_cell_data_func (column,
+	                                         cell_renderer,
+	                                         (GtkTreeCellDataFunc)render_proposal_accelerator_func,
+	                                         completion,
+	                                         NULL);
+
+	g_signal_connect_object (completion,
+				 "notify::accelerators",
+				 G_CALLBACK (accelerators_notify_cb),
+				 column,
+				 0);
+}
+
+static void
+init_main_window (GtkSourceCompletion *completion,
+		  GtkBuilder          *builder)
+{
+	if (completion->priv->view == NULL)
+	{
+		return;
+	}
+
+	completion->priv->main_window = GTK_SOURCE_COMPLETION_INFO (gtk_builder_get_object (builder, "main_window"));
+	completion->priv->info_button = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "info_button"));
+	completion->priv->selection_image = GTK_IMAGE (gtk_builder_get_object (builder, "selection_image"));
+	completion->priv->selection_label = GTK_LABEL (gtk_builder_get_object (builder, "selection_label"));
+	completion->priv->bottom_bar = GTK_WIDGET (gtk_builder_get_object (builder, "bottom_bar"));
+
+	gtk_container_set_border_width (GTK_CONTAINER (completion->priv->main_window), 0);
+
+	gtk_window_set_attached_to (GTK_WINDOW (completion->priv->main_window),
+				    GTK_WIDGET (completion->priv->view));
+
+	g_signal_connect (completion->priv->main_window,
+			  "configure-event",
+			  G_CALLBACK (gtk_source_completion_configure_event),
+			  completion);
+
+	g_signal_connect_swapped (completion->priv->main_window,
+				  "size-allocate",
+				  G_CALLBACK (update_window_position),
+				  completion);
+
+	g_signal_connect (completion->priv->main_window,
+			  "delete-event",
+			  G_CALLBACK (gtk_widget_hide_on_delete),
+			  NULL);
+
+	g_signal_connect (completion->priv->main_window,
+	                  "notify::transient-for",
+	                  G_CALLBACK (update_transient_for_info),
+	                  completion);
+
+	g_signal_connect_swapped (completion->priv->info_button,
+				  "toggled",
+				  G_CALLBACK (update_info_window_visibility),
+				  completion);
+}
+
+static void
+init_info_window (GtkSourceCompletion *completion)
+{
+	completion->priv->info_window = gtk_source_completion_info_new ();
+	g_object_ref_sink (completion->priv->info_window);
+
+	gtk_window_set_attached_to (GTK_WINDOW (completion->priv->info_window),
+				    GTK_WIDGET (completion->priv->main_window));
+
+	g_signal_connect_swapped (completion->priv->info_window,
+				  "size-allocate",
+				  G_CALLBACK (update_info_position),
+				  completion);
+
+	/* Default info widget */
+
+	completion->priv->default_info = GTK_LABEL (gtk_label_new (NULL));
+	g_object_ref_sink (completion->priv->default_info);
+
+	gtk_widget_show (GTK_WIDGET (completion->priv->default_info));
+}
+
+static void
+connect_style_context (GtkSourceCompletion *completion)
+{
+	GtkStyleContext *style_context;
+
+	if (completion->priv->view == NULL)
+	{
+		return;
+	}
+
+	style_context = gtk_widget_get_style_context (GTK_WIDGET (completion->priv->view));
+
+	g_signal_connect_object (style_context,
+				 "changed",
+				 G_CALLBACK (style_context_changed),
+				 completion,
+				 G_CONNECT_AFTER);
+
+	style_context_changed (style_context, completion);
+}
+
+static void
+gtk_source_completion_constructed (GObject *object)
+{
+	GtkSourceCompletion *completion = GTK_SOURCE_COMPLETION (object);
+	GError *error = NULL;
+	GtkBuilder *builder = gtk_builder_new ();
+	GtkSourceCompletionContainer *container = _gtk_source_completion_container_new ();
+	g_object_ref_sink (container);
+
+	gtk_builder_set_translation_domain (builder, GETTEXT_PACKAGE);
+
+	/* GtkSourceCompletionContainer is a private type. */
+	gtk_builder_expose_object (builder, "completion_container", G_OBJECT (container));
+
+	gtk_builder_add_from_resource (builder,
+				       "/org/gnome/gtksourceview/ui/gtksourcecompletion.ui",
+				       &error);
+
+	if (error != NULL)
+	{
+		g_error ("Error while loading the completion UI: %s", error->message);
+	}
+
+	init_tree_view (completion, builder);
+	init_main_window (completion, builder);
+	init_info_window (completion);
+	connect_style_context (completion);
+
+	g_object_unref (builder);
+	g_object_unref (container);
+
+	G_OBJECT_CLASS (gtk_source_completion_parent_class)->constructed (object);
+}
+
 static void
 gtk_source_completion_class_init (GtkSourceCompletionClass *klass)
 {
@@ -1782,6 +2167,7 @@ gtk_source_completion_class_init (GtkSourceCompletionClass *klass)
 	object_class->get_property = gtk_source_completion_get_property;
 	object_class->set_property = gtk_source_completion_set_property;
 	object_class->dispose = gtk_source_completion_dispose;
+	object_class->constructed = gtk_source_completion_constructed;
 
 	klass->show = gtk_source_completion_show_default;
 	klass->hide = gtk_source_completion_hide_default;
@@ -1942,11 +2328,8 @@ gtk_source_completion_class_init (GtkSourceCompletionClass *klass)
 			      G_TYPE_FROM_CLASS (klass),
 			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
 			      G_STRUCT_OFFSET (GtkSourceCompletionClass, show),
-			      NULL,
-			      NULL,
-			      g_cclosure_marshal_VOID__VOID,
-			      G_TYPE_NONE,
-			      0);
+			      NULL, NULL, NULL,
+			      G_TYPE_NONE, 0);
 
 
 	/**
@@ -1961,11 +2344,8 @@ gtk_source_completion_class_init (GtkSourceCompletionClass *klass)
 			      G_TYPE_FROM_CLASS (klass),
 			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
 			      G_STRUCT_OFFSET (GtkSourceCompletionClass, hide),
-			      NULL,
-			      NULL,
-			      g_cclosure_marshal_VOID__VOID,
-			      G_TYPE_NONE,
-			      0);
+			      NULL, NULL, NULL,
+			      G_TYPE_NONE, 0);
 
 	/**
 	 * GtkSourceCompletion::populate-context:
@@ -1980,9 +2360,7 @@ gtk_source_completion_class_init (GtkSourceCompletionClass *klass)
 		              G_TYPE_FROM_CLASS (klass),
 		              G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
 		              G_STRUCT_OFFSET (GtkSourceCompletionClass, populate_context),
-		              NULL,
-		              NULL,
-		              g_cclosure_marshal_VOID__OBJECT,
+		              NULL, NULL, NULL,
 		              G_TYPE_NONE,
 		              1,
 		              GTK_SOURCE_TYPE_COMPLETION_CONTEXT);
@@ -2004,7 +2382,7 @@ gtk_source_completion_class_init (GtkSourceCompletionClass *klass)
 	 * <keycap>Home</keycap> and <keycap>End</keycap> keys are bound to the
 	 * normal behavior expected by those keys.
 	 *
-	 * When @step is equal to #GTK_SCROLL_PAGES, the page size is defined by
+	 * When @step is equal to %GTK_SCROLL_PAGES, the page size is defined by
 	 * the #GtkSourceCompletion:proposal-page-size property. It is used for
 	 * the <keycap>PageDown</keycap> and <keycap>PageUp</keycap> keys.
 	 *
@@ -2017,9 +2395,7 @@ gtk_source_completion_class_init (GtkSourceCompletionClass *klass)
 			      G_TYPE_FROM_CLASS (klass),
 			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
 			      G_STRUCT_OFFSET (GtkSourceCompletionClass, move_cursor),
-			      NULL,
-			      NULL,
-			      _gtksourceview_marshal_VOID__ENUM_INT,
+			      NULL, NULL, NULL,
 			      G_TYPE_NONE,
 			      2,
 			      GTK_TYPE_SCROLL_STEP,
@@ -2056,9 +2432,7 @@ gtk_source_completion_class_init (GtkSourceCompletionClass *klass)
 			      G_TYPE_FROM_CLASS (klass),
 			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
 			      G_STRUCT_OFFSET (GtkSourceCompletionClass, move_page),
-			      NULL,
-			      NULL,
-			      _gtksourceview_marshal_VOID__ENUM_INT,
+			      NULL, NULL, NULL,
 			      G_TYPE_NONE,
 			      2,
 			      GTK_TYPE_SCROLL_STEP,
@@ -2081,11 +2455,8 @@ gtk_source_completion_class_init (GtkSourceCompletionClass *klass)
 			      G_TYPE_FROM_CLASS (klass),
 			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
 			      G_STRUCT_OFFSET (GtkSourceCompletionClass, activate_proposal),
-			      NULL,
-			      NULL,
-			      g_cclosure_marshal_VOID__VOID,
-			      G_TYPE_NONE,
-			      0);
+			      NULL, NULL, NULL,
+			      G_TYPE_NONE, 0);
 
 	/* Key bindings */
 	binding_set = gtk_binding_set_by_class (klass);
@@ -2189,241 +2560,31 @@ gtk_source_completion_class_init (GtkSourceCompletionClass *klass)
 				      G_TYPE_INT, 1);
 }
 
-static gboolean
-selection_func (GtkTreeSelection    *selection,
-                GtkTreeModel        *model,
-                GtkTreePath         *path,
-                gboolean             path_currently_selected,
-                GtkSourceCompletion *completion)
-{
-	GtkTreeIter iter;
-
-	gtk_tree_model_get_iter (model, &iter, path);
-
-	if (gtk_source_completion_model_iter_is_header (completion->priv->model_proposals,
-	                                                &iter))
-	{
-		/* A header must never be selected */
-		g_return_val_if_fail (!path_currently_selected, TRUE);
-		return FALSE;
-	}
-	else
-	{
-		return TRUE;
-	}
-}
-
-static void
-accelerators_notify_cb (GtkSourceCompletion *completion,
-			GParamSpec          *pspec,
-			GtkTreeViewColumn   *column)
-{
-	gtk_tree_view_column_set_visible (column, completion->priv->num_accelerators > 0);
-}
-
-static void
-init_tree_view (GtkSourceCompletion *completion,
-		GtkBuilder          *builder)
-{
-	GtkTreeSelection *selection;
-	GtkTreeViewColumn *column;
-	GtkCellRenderer *cell_renderer;
-	GtkStyleContext *style_context;
-	GdkRGBA background_color;
-	GdkRGBA foreground_color;
-
-	completion->priv->tree_view_proposals = GTK_TREE_VIEW (gtk_builder_get_object (builder, "tree_view_proposals"));
-
-	g_signal_connect_swapped (completion->priv->tree_view_proposals,
-				  "row-activated",
-				  G_CALLBACK (gtk_source_completion_activate_proposal),
-				  completion);
-
-	g_signal_connect_swapped (completion->priv->tree_view_proposals,
-				  "size-allocate",
-				  G_CALLBACK (gtk_source_completion_proposals_size_allocate),
-				  completion);
-
-	/* Selection */
-
-	selection = gtk_tree_view_get_selection (completion->priv->tree_view_proposals);
-
-	gtk_tree_selection_set_select_function (selection,
-	                                        (GtkTreeSelectionFunc)selection_func,
-	                                        completion,
-	                                        NULL);
-
-	g_signal_connect (selection,
-			  "changed",
-			  G_CALLBACK (selection_changed_cb),
-			  completion);
-
-	/* Icon cell renderer */
-
-	cell_renderer = GTK_CELL_RENDERER (gtk_builder_get_object (builder, "cell_renderer_icon"));
-
-	column = GTK_TREE_VIEW_COLUMN (gtk_builder_get_object (builder, "tree_view_column_icon"));
-
-	gtk_tree_view_column_set_attributes (column, cell_renderer,
-					     "pixbuf", GTK_SOURCE_COMPLETION_MODEL_COLUMN_ICON,
-					     "cell-background-set", GTK_SOURCE_COMPLETION_MODEL_COLUMN_IS_HEADER,
-					     NULL);
-
-	style_context = gtk_widget_get_style_context (GTK_WIDGET (completion->priv->tree_view_proposals));
-	gtk_style_context_get_background_color (style_context,
-	                                        GTK_STATE_FLAG_INSENSITIVE,
-	                                        &background_color);
-
-	g_object_set (cell_renderer,
-	              "cell-background-rgba", &background_color,
-	              NULL);
-
-	g_object_bind_property (completion, "show-icons",
-				cell_renderer, "visible",
-				G_BINDING_SYNC_CREATE);
-
-	/* Proposal text cell renderer */
-
-	cell_renderer = GTK_CELL_RENDERER (gtk_builder_get_object (builder, "cell_renderer_proposal"));
-	completion->priv->cell_renderer_proposal = cell_renderer;
-
-	column = GTK_TREE_VIEW_COLUMN (gtk_builder_get_object (builder, "tree_view_column_proposal"));
-
-	gtk_tree_view_column_set_attributes (column, cell_renderer,
-					     "markup", GTK_SOURCE_COMPLETION_MODEL_COLUMN_MARKUP,
-					     "cell-background-set", GTK_SOURCE_COMPLETION_MODEL_COLUMN_IS_HEADER,
-					     "foreground-set", GTK_SOURCE_COMPLETION_MODEL_COLUMN_IS_HEADER,
-					     NULL);
-
-	gtk_style_context_get_color (style_context,
-	                             GTK_STATE_FLAG_INSENSITIVE,
-	                             &foreground_color);
-
-	g_object_set (cell_renderer,
-	              "foreground-rgba", &foreground_color,
-	              "cell-background-rgba", &background_color,
-	              NULL);
-
-	/* Accelerators cell renderer */
-
-	column = GTK_TREE_VIEW_COLUMN (gtk_builder_get_object (builder, "tree_view_column_accelerator"));
-
-	cell_renderer = GTK_CELL_RENDERER (gtk_builder_get_object (builder, "cell_renderer_accelerator"));
-
-	gtk_tree_view_column_set_attributes (column,
-					     cell_renderer,
-					     "cell-background-set", GTK_SOURCE_COMPLETION_MODEL_COLUMN_IS_HEADER,
-					     NULL);
-
-	g_object_set (cell_renderer,
-	              "foreground-rgba", &foreground_color,
-	              "cell-background-rgba", &background_color,
-		      NULL);
-
-	gtk_tree_view_column_set_cell_data_func (column,
-	                                         cell_renderer,
-	                                         (GtkTreeCellDataFunc)render_proposal_accelerator_func,
-	                                         completion,
-	                                         NULL);
-
-	g_signal_connect_object (completion,
-				 "notify::accelerators",
-				 G_CALLBACK (accelerators_notify_cb),
-				 column,
-				 0);
-}
-
-static void
-init_main_window (GtkSourceCompletion *completion,
-		  GtkBuilder          *builder)
-{
-	completion->priv->main_window = GTK_SOURCE_COMPLETION_INFO (gtk_builder_get_object (builder, "main_window"));
-	completion->priv->info_button = GTK_TOGGLE_BUTTON (gtk_builder_get_object (builder, "info_button"));
-	completion->priv->selection_image = GTK_IMAGE (gtk_builder_get_object (builder, "selection_image"));
-	completion->priv->selection_label = GTK_LABEL (gtk_builder_get_object (builder, "selection_label"));
-	completion->priv->bottom_bar = GTK_WIDGET (gtk_builder_get_object (builder, "bottom_bar"));
-
-	gtk_window_set_attached_to (GTK_WINDOW (completion->priv->main_window),
-				    GTK_WIDGET (completion->priv->view));
-
-	g_signal_connect (completion->priv->main_window,
-			  "configure-event",
-			  G_CALLBACK (gtk_source_completion_configure_event),
-			  completion);
-
-	g_signal_connect_swapped (completion->priv->main_window,
-				  "size-allocate",
-				  G_CALLBACK (update_window_position),
-				  completion);
-
-	g_signal_connect (completion->priv->main_window,
-			  "delete-event",
-			  G_CALLBACK (gtk_widget_hide_on_delete),
-			  NULL);
-
-	g_signal_connect (completion->priv->main_window,
-	                  "notify::transient-for",
-	                  G_CALLBACK (update_transient_for_info),
-	                  completion);
-
-	g_signal_connect_swapped (completion->priv->info_button,
-				  "toggled",
-				  G_CALLBACK (update_info_window_visibility),
-				  completion);
-}
-
-static void
-init_info_window (GtkSourceCompletion *completion)
-{
-	completion->priv->info_window = gtk_source_completion_info_new ();
-	g_object_ref_sink (completion->priv->info_window);
-
-	gtk_window_set_attached_to (GTK_WINDOW (completion->priv->info_window),
-				    GTK_WIDGET (completion->priv->main_window));
-
-	g_signal_connect_swapped (completion->priv->info_window,
-				  "size-allocate",
-				  G_CALLBACK (update_info_position),
-				  completion);
-
-	/* Default info widget */
-
-	completion->priv->default_info = GTK_LABEL (gtk_label_new (NULL));
-	g_object_ref_sink (completion->priv->default_info);
-
-	gtk_widget_show (GTK_WIDGET (completion->priv->default_info));
-}
-
 static void
 gtk_source_completion_init (GtkSourceCompletion *completion)
 {
-	GError *error = NULL;
-	GtkBuilder *builder = gtk_builder_new ();
-	GtkSourceCompletionContainer *container = _gtk_source_completion_container_new ();
-	g_object_ref_sink (container);
-
 	completion->priv = gtk_source_completion_get_instance_private (completion);
+}
 
-	gtk_builder_set_translation_domain (builder, GETTEXT_PACKAGE);
+static GObject *
+gtk_source_completion_buildable_get_internal_child (GtkBuildable *buildable,
+						    GtkBuilder   *builder,
+						    const gchar  *childname)
+{
+	GtkSourceCompletion *completion = GTK_SOURCE_COMPLETION (buildable);
 
-	/* GtkSourceCompletionContainer is a private type. */
-	gtk_builder_expose_object (builder, "completion_container", G_OBJECT (container));
-
-	gtk_builder_add_from_resource (builder,
-				       "/org/gnome/gtksourceview/ui/gtksourcecompletion.ui",
-				       &error);
-
-	if (error != NULL)
+	if (g_strcmp0 (childname, "info_window") == 0)
 	{
-		g_error ("Error while loading the completion UI: %s", error->message);
+		return G_OBJECT (gtk_source_completion_get_info_window (completion));
 	}
 
-	init_tree_view (completion, builder);
-	init_main_window (completion, builder);
-	init_info_window (completion);
+	return NULL;
+}
 
-	g_object_unref (builder);
-	g_object_unref (container);
+static void
+gtk_source_completion_buildable_interface_init (GtkBuildableIface *iface)
+{
+	iface->get_internal_child = gtk_source_completion_buildable_get_internal_child;
 }
 
 void
@@ -2491,6 +2652,11 @@ gtk_source_completion_show (GtkSourceCompletion        *completion,
 	g_return_val_if_fail (GTK_SOURCE_IS_COMPLETION (completion), FALSE);
 	g_return_val_if_fail (GTK_SOURCE_IS_COMPLETION_CONTEXT (context), FALSE);
 
+	if (completion->priv->view == NULL)
+	{
+		return FALSE;
+	}
+
 	/* Make sure to clear any active completion */
 	reset_completion (completion);
 
@@ -2707,10 +2873,11 @@ gtk_source_completion_get_info_window (GtkSourceCompletion *completion)
  * gtk_source_completion_get_view:
  * @completion: a #GtkSourceCompletion.
  *
- * The #GtkSourceView associated with @completion.
+ * The #GtkSourceView associated with @completion, or %NULL if the view has been
+ * destroyed.
  *
  * Returns: (type GtkSource.View) (transfer none):
- * The #GtkSourceView associated with @completion.
+ * The #GtkSourceView associated with @completion, or %NULL.
  */
 GtkSourceView *
 gtk_source_completion_get_view (GtkSourceCompletion *completion)
@@ -2742,6 +2909,11 @@ gtk_source_completion_create_context (GtkSourceCompletion *completion,
 
 	g_return_val_if_fail (GTK_SOURCE_IS_COMPLETION (completion), NULL);
 
+	if (completion->priv->view == NULL)
+	{
+		return NULL;
+	}
+
 	if (position == NULL)
 	{
 		get_iter_at_insert (completion, &iter);
@@ -2770,6 +2942,11 @@ gtk_source_completion_move_window (GtkSourceCompletion *completion,
 	g_return_if_fail (GTK_SOURCE_IS_COMPLETION (completion));
 	g_return_if_fail (iter != NULL);
 
+	if (completion->priv->view == NULL)
+	{
+		return;
+	}
+
 	if (!gtk_widget_get_visible (GTK_WIDGET (completion->priv->main_window)))
 	{
 		return;
@@ -2788,18 +2965,27 @@ gtk_source_completion_move_window (GtkSourceCompletion *completion,
  * completion when inserting or deleting text from the buffer associated with
  * the completion. Use gtk_source_completion_unblock_interactive() to enable
  * interactive completion again.
+ *
+ * This function may be called multiple times. It will continue to block
+ * interactive completion until gtk_source_completion_unblock_interactive()
+ * has been called the same number of times.
  */
 void
 gtk_source_completion_block_interactive (GtkSourceCompletion *completion)
 {
-	GtkTextBuffer *buffer;
-
 	g_return_if_fail (GTK_SOURCE_IS_COMPLETION (completion));
 
-	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (completion->priv->view));
+	if (completion->priv->view == NULL)
+	{
+		return;
+	}
 
-	g_signal_handler_block (buffer, completion->priv->signals_ids[TEXT_BUFFER_INSERT_TEXT]);
-	g_signal_handler_block (buffer, completion->priv->signals_ids[TEXT_BUFFER_DELETE_RANGE]);
+	if (completion->priv->block_interactive_num == 0)
+	{
+		block_interactive (completion);
+	}
+
+	completion->priv->block_interactive_num++;
 }
 
 /**
@@ -2813,12 +2999,26 @@ gtk_source_completion_block_interactive (GtkSourceCompletion *completion)
 void
 gtk_source_completion_unblock_interactive (GtkSourceCompletion *completion)
 {
-	GtkTextBuffer *buffer;
-
 	g_return_if_fail (GTK_SOURCE_IS_COMPLETION (completion));
 
-	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (completion->priv->view));
+	if (completion->priv->view == NULL)
+	{
+		return;
+	}
 
-	g_signal_handler_unblock (buffer, completion->priv->signals_ids[TEXT_BUFFER_INSERT_TEXT]);
-	g_signal_handler_unblock (buffer, completion->priv->signals_ids[TEXT_BUFFER_DELETE_RANGE]);
+	if (completion->priv->block_interactive_num == 1)
+	{
+		g_signal_handlers_unblock_by_func (completion->priv->buffer,
+						   buffer_insert_text_cb,
+						   completion);
+
+		g_signal_handlers_unblock_by_func (completion->priv->buffer,
+						   buffer_delete_range_cb,
+						   completion);
+	}
+
+	if (completion->priv->block_interactive_num > 0)
+	{
+		completion->priv->block_interactive_num--;
+	}
 }
