@@ -30,7 +30,7 @@
 
 /**
  * SECTION:stylescheme
- * @Short_description: Object controlling the appearance of #GtkSourceView
+ * @Short_description: Object controlling the appearance of GtkSourceView
  * @Title: GtkSourceStyleScheme
  * @See_also: #GtkSourceStyle, #GtkSourceStyleSchemeManager
  *
@@ -317,7 +317,7 @@ gtk_source_style_scheme_get_name (GtkSourceStyleScheme *scheme)
  * gtk_source_style_scheme_get_description:
  * @scheme: a #GtkSourceStyleScheme.
  *
- * Returns: @scheme description (if defined), or %NULL.
+ * Returns: (nullable): @scheme description (if defined), or %NULL.
  *
  * Since: 2.0
  */
@@ -333,9 +333,9 @@ gtk_source_style_scheme_get_description (GtkSourceStyleScheme *scheme)
  * gtk_source_style_scheme_get_authors:
  * @scheme: a #GtkSourceStyleScheme.
  *
- * Returns: (array zero-terminated=1) (transfer none): a %NULL-terminated
- * array containing the @scheme authors or %NULL if no author
- * is specified by the style scheme.
+ * Returns: (nullable) (array zero-terminated=1) (transfer none): a
+ * %NULL-terminated array containing the @scheme authors or %NULL if
+ * no author is specified by the style scheme.
  *
  * Since: 2.0
  */
@@ -356,8 +356,8 @@ gtk_source_style_scheme_get_authors (GtkSourceStyleScheme *scheme)
  * gtk_source_style_scheme_get_filename:
  * @scheme: a #GtkSourceStyleScheme.
  *
- * Returns: @scheme file name if the scheme was created parsing a
- * style scheme file or %NULL in the other cases.
+ * Returns: (nullable): @scheme file name if the scheme was created
+ * parsing a style scheme file or %NULL in the other cases.
  *
  * Since: 2.0
  */
@@ -496,9 +496,9 @@ fix_style_colors (GtkSourceStyleScheme *scheme,
  * @scheme: a #GtkSourceStyleScheme.
  * @style_id: id of the style to retrieve.
  *
- * Returns: (transfer none): style which corresponds to @style_id
- * in the @scheme, or %NULL when no style with this name found.
- * It is owned by @scheme and may not be unref'ed.
+ * Returns: (nullable) (transfer none): style which corresponds to @style_id in
+ * the @scheme, or %NULL when no style with this name found.  It is owned by
+ * @scheme and may not be unref'ed.
  *
  * Since: 2.0
  */
@@ -635,33 +635,6 @@ _gtk_source_style_scheme_get_current_line_color (GtkSourceStyleScheme *scheme,
 	return get_color (style, FALSE, color);
 }
 
-static void
-set_line_numbers_style (GtkWidget      *widget,
-			GtkSourceStyle *style)
-{
-	GdkRGBA *fg_ptr = NULL;
-	GdkRGBA *bg_ptr = NULL;
-	GdkRGBA fg;
-	GdkRGBA bg;
-	GtkStateFlags flags;
-
-	if (get_color (style, TRUE, &fg))
-	{
-		fg_ptr = &fg;
-	}
-
-	if (get_color (style, FALSE, &bg))
-	{
-		bg_ptr = &bg;
-	}
-
-	/* Override the color no matter what the state is */
-	flags = GTK_STATE_FLAG_NORMAL;
-
-	gtk_widget_override_color (widget, flags, fg_ptr);
-	gtk_widget_override_background_color (widget, flags, bg_ptr);
-}
-
 static void
 update_cursor_colors (GtkWidget      *widget,
 		      GtkSourceStyle *style_primary,
@@ -733,10 +706,10 @@ _gtk_source_style_scheme_apply (GtkSourceStyleScheme *scheme,
 	                                GTK_STYLE_PROVIDER (scheme->priv->css),
 	                                GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
 
+	G_GNUC_BEGIN_IGNORE_DEPRECATIONS;
+	/* See https://bugzilla.gnome.org/show_bug.cgi?id=708583 */
 	gtk_style_context_invalidate (context);
-
-	style = gtk_source_style_scheme_get_style (scheme, STYLE_LINE_NUMBERS);
-	set_line_numbers_style (widget, style);
+	G_GNUC_END_IGNORE_DEPRECATIONS;
 
 	style = gtk_source_style_scheme_get_style (scheme, STYLE_CURSOR);
 	style2 = gtk_source_style_scheme_get_style (scheme, STYLE_SECONDARY_CURSOR);
@@ -764,8 +737,12 @@ _gtk_source_style_scheme_unapply (GtkSourceStyleScheme *scheme,
 	context = gtk_widget_get_style_context (GTK_WIDGET (widget));
 	gtk_style_context_remove_provider (context,
 	                                   GTK_STYLE_PROVIDER (scheme->priv->css));
+
+	G_GNUC_BEGIN_IGNORE_DEPRECATIONS;
+	/* See https://bugzilla.gnome.org/show_bug.cgi?id=708583 */
 	gtk_style_context_invalidate (context);
-	set_line_numbers_style (widget, NULL);
+	G_GNUC_END_IGNORE_DEPRECATIONS;
+
 	update_cursor_colors (widget, NULL, NULL);
 }
 
@@ -784,7 +761,7 @@ get_css_color_style (GtkSourceStyle *style,
 	{
 		gchar *bg_color;
 		bg_color = gdk_rgba_to_string (&color);
-		*bg = g_strdup_printf ("%s: %s;\n", "background-color", bg_color);
+		*bg = g_strdup_printf ("background-color: %s;\n", bg_color);
 		g_free (bg_color);
 	}
 	else
@@ -796,7 +773,7 @@ get_css_color_style (GtkSourceStyle *style,
 	{
 		gchar *text_color;
 		text_color = gdk_rgba_to_string (&color);
-		*text = g_strdup_printf ("%s: %s;\n", "color", text_color);
+		*text = g_strdup_printf ("color: %s;\n", text_color);
 		g_free (text_color);
 	}
 	else
@@ -808,20 +785,19 @@ get_css_color_style (GtkSourceStyle *style,
 static void
 append_css_style (GString        *string,
                   GtkSourceStyle *style,
-                  const gchar    *state)
+                  const gchar    *selector)
 {
 	gchar *bg, *text;
 	const gchar css_style[] =
-		".view%s {\n"
+		"%s {\n"
 		"	%s"
 		"	%s"
 		"}\n";
 
 	get_css_color_style (style, &bg, &text);
-
 	if (bg || text)
 	{
-		g_string_append_printf (string, css_style, state,
+		g_string_append_printf (string, css_style, selector,
 		                        bg != NULL ? bg : "",
 		                        text != NULL ? text : "");
 
@@ -839,15 +815,30 @@ generate_css_style (GtkSourceStyleScheme *scheme)
 	final_style = g_string_new ("");
 
 	style = gtk_source_style_scheme_get_style (scheme, STYLE_TEXT);
-	append_css_style (final_style, style, "");
+	append_css_style (final_style, style, ".view");
 
 	style = gtk_source_style_scheme_get_style (scheme, STYLE_SELECTED);
-	append_css_style (final_style, style, ":selected:focused");
+	append_css_style (final_style, style, ".view:selected:focused");
 
 	style2 = gtk_source_style_scheme_get_style (scheme, STYLE_SELECTED_UNFOCUSED);
 	if (style2 == NULL)
 		style2 = style;
-	append_css_style (final_style, style2, ":selected");
+	append_css_style (final_style, style2, ".view:selected");
+
+	/* For now we use "line numbers" colors for all the gutters */
+	style = gtk_source_style_scheme_get_style (scheme, STYLE_LINE_NUMBERS);
+	if (style != NULL)
+	{
+		append_css_style (final_style, style, ".top");
+		append_css_style (final_style, style, ".right");
+		append_css_style (final_style, style, ".bottom");
+		append_css_style (final_style, style, ".left");
+
+		/* For the corners if the top or bottom gutter is also
+		 * displayed.
+		 */
+		append_css_style (final_style, style, "GtkSourceView");
+	}
 
 	if (*final_style->str != '\0')
 	{
@@ -902,6 +893,7 @@ parse_style (GtkSourceStyleScheme *scheme,
 	gboolean italic = FALSE;
 	gboolean underline = FALSE;
 	gboolean strikethrough = FALSE;
+	xmlChar *scale = NULL;
 	xmlChar *tmp;
 
 	tmp = xmlGetProp (node, BAD_CAST "name");
@@ -948,10 +940,11 @@ parse_style (GtkSourceStyleScheme *scheme,
 	get_bool (node, "bold", &mask, GTK_SOURCE_STYLE_USE_BOLD, &bold);
 	get_bool (node, "underline", &mask, GTK_SOURCE_STYLE_USE_UNDERLINE, &underline);
 	get_bool (node, "strikethrough", &mask, GTK_SOURCE_STYLE_USE_STRIKETHROUGH, &strikethrough);
+	scale = xmlGetProp (node, BAD_CAST "scale");
 
 	if (use_style)
 	{
-		if (fg != NULL || bg != NULL || line_bg != NULL || mask != 0)
+		if (fg != NULL || bg != NULL || line_bg != NULL || mask != 0 || scale != NULL)
 		{
 			g_set_error (error, ERROR_QUARK, 0,
 				     "in style '%s': style attributes used along with use-style",
@@ -994,11 +987,17 @@ parse_style (GtkSourceStyleScheme *scheme,
 			result->line_background = g_intern_string ((char*) line_bg);
 			result->mask |= GTK_SOURCE_STYLE_USE_LINE_BACKGROUND;
 		}
+		if (scale != NULL)
+		{
+			result->scale = g_intern_string ((char*) scale);
+			result->mask |= GTK_SOURCE_STYLE_USE_SCALE;
+		}
 	}
 
 	*style_p = result;
 	*style_name_p = style_name;
 
+	xmlFree (scale);
 	xmlFree (line_bg);
 	xmlFree (bg);
 	xmlFree (fg);
@@ -1160,8 +1159,8 @@ parse_style_scheme_element (GtkSourceStyleScheme *scheme,
  * _gtk_source_style_scheme_new_from_file:
  * @filename: file to parse.
  *
- * Returns: new #GtkSourceStyleScheme created from file, or
- * %NULL on error.
+ * Returns: (nullable): new #GtkSourceStyleScheme created from file,
+ * or %NULL on error.
  *
  * Since: 2.0
  */
@@ -1241,7 +1240,7 @@ _gtk_source_style_scheme_new_from_file (const gchar *filename)
  * _gtk_source_style_scheme_get_parent_id:
  * @scheme: a #GtkSourceStyleScheme.
  *
- * Returns: parent style scheme id or %NULL.
+ * Returns: (nullable): parent style scheme id or %NULL.
  *
  * Since: 2.0
  */
